@@ -322,11 +322,8 @@ BTRFS_SETGET_FUNCS(timespec_nsec, struct btrfs_timespec, nsec, 32);
 BTRFS_SETGET_STACK_FUNCS(stack_timespec_sec, struct btrfs_timespec, sec, 64);
 BTRFS_SETGET_STACK_FUNCS(stack_timespec_nsec, struct btrfs_timespec, nsec, 32);
 
-BTRFS_SETGET_FUNCS(stripe_extent_encoding, struct btrfs_stripe_extent, encoding, 8);
 BTRFS_SETGET_FUNCS(raid_stride_devid, struct btrfs_raid_stride, devid, 64);
 BTRFS_SETGET_FUNCS(raid_stride_offset, struct btrfs_raid_stride, offset, 64);
-BTRFS_SETGET_STACK_FUNCS(stack_stripe_extent_encoding,
-			 struct btrfs_stripe_extent, encoding, 8);
 BTRFS_SETGET_STACK_FUNCS(stack_raid_stride_devid, struct btrfs_raid_stride, devid, 64);
 
 static inline struct btrfs_raid_stride *btrfs_raid_stride_nr(
@@ -670,17 +667,17 @@ static inline void btrfs_dir_item_key_to_cpu(const struct extent_buffer *eb,
 static inline void btrfs_disk_key_to_cpu(struct btrfs_key *cpu,
 					 const struct btrfs_disk_key *disk)
 {
-	cpu->offset = le64_to_cpu(disk->offset);
-	cpu->type = disk->type;
 	cpu->objectid = le64_to_cpu(disk->objectid);
+	cpu->type = disk->type;
+	cpu->offset = le64_to_cpu(disk->offset);
 }
 
 static inline void btrfs_cpu_key_to_disk(struct btrfs_disk_key *disk,
 					 const struct btrfs_key *cpu)
 {
-	disk->offset = cpu_to_le64(cpu->offset);
-	disk->type = cpu->type;
 	disk->objectid = cpu_to_le64(cpu->objectid);
+	disk->type = cpu->type;
+	disk->offset = cpu_to_le64(cpu->offset);
 }
 
 static inline void btrfs_node_key_to_cpu(const struct extent_buffer *eb,
